package loopbounds.parsetree;

import java.util.List;
import java.util.ArrayList;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.TreeMaker;
import com.sun.tools.javac.tree.JCTree.*;
import com.sun.tools.javac.util.Context;

public class DNFSplitter {
	private ConditionTransformer ct;
	private Context context;
	
	/** Constructor.
	 * 
	 * @param code The parse tree for which to create the splitter.
	 */
	public DNFSplitter(ParseTreeWrapper code) {
		ct = new ConditionTransformer(code.context());
		context = code.context();
	}
	
	/** Transform a condition into Disjunctional Normal Form.
	 * 
	 * @param condition The condition to transform
	 * @return The condition in DNF
	 */
	public JCExpression DNF(JCExpression condition) {
		return ct.DNF(condition);
	}
	
	/** Split up a loop into multiple loop with conjunctive exit-conditions,
	 * i.e. perform DNF-splitting.
	 * 
	 * Puts the loops condition in DNF, then makes new loops, where each has a condition
	 * which is one of the conjunctive pieces of the disjunctive exit-condition.
	 * 
	 * Example: splits up a loop with condition (i < 10) || (i > 30) into three loops,
	 * with respective conditions (i < 10), (i > 30) and (i < 10) && (i > 30).
	 * 
	 * @param loop The loop to split up
	 * @return A list of loops
	 */
	public List<JCStatement> split(JCStatement loop) {
		List<JCStatement> loops = new ArrayList<JCStatement>();
		
		JCExpression cond = ParseTreeWrapper.getLoopCondition(loop);
		if(cond == null){
			TreeMaker maker = TreeMaker.instance(context);
			cond = maker.Literal(true);
		}			
		JCExpression condDNF = ct.DNF((JCExpression) cond.clone());
		List<JCTree.JCExpression> condlist = ct.splitOR(condDNF);
		
		for (JCExpression newcond : condlist) {
			JCStatement newloop = (JCStatement)loop.clone();
			ParseTreeWrapper.setLoopCondition(newloop, newcond);
			loops.add(newloop);
		}
		return loops;
	}
}
