package loopbounds.parsetree;

import java.util.List;
import java.util.ArrayList;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;
import org.jmlspecs.openjml.JmlTree.JmlMethodDecl;

import com.sun.tools.javac.tree.JCTree;

/**
 * Class for retrieving all method declarations in a JmlCompilationUnit. It is
 * intended to be used through the static function 'getMethods(JCTree tree)'
 * which takes a tree node as an argument and returns a list of JmlMethodDecl's.
 * <P>
 * The class extends JmlTreeScanner and uses the visitor pattern.
 * 
 * @author Rody Kersten
 * @author Martijn Moraal
 *
 */
public class MethodScanner extends JmlTreeScanner implements IJmlVisitor {
	
	private List<JmlMethodDecl> methods = new ArrayList<JmlMethodDecl>();
	
	/** Get the methods in a parse tree.
	 * 
	 * @param tree The tree to look in
	 * @return A list of the methods in 'tree'
	 */
	public static List<JmlMethodDecl> getMethods (JCTree tree) {
		MethodScanner scanner = new MethodScanner();
		scanner.scan(tree);
		return scanner.methods;
	}
	
	@Override
	public void visitJmlMethodDecl (JmlMethodDecl tree) {
		if (tree.body != null)
			methods.add(tree);
	}
}
