package loopbounds.parsetree;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeTranslator;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCArrayAccess;
import com.sun.tools.javac.tree.JCTree.JCAssign;
import com.sun.tools.javac.tree.JCTree.JCAssignOp;
import com.sun.tools.javac.tree.JCTree.JCBinary;
import com.sun.tools.javac.tree.JCTree.JCExpression;
import com.sun.tools.javac.tree.JCTree.JCFieldAccess;
import com.sun.tools.javac.tree.JCTree.JCIdent;
import com.sun.tools.javac.tree.JCTree.JCIf;
import com.sun.tools.javac.tree.JCTree.JCMethodInvocation;
import com.sun.tools.javac.tree.JCTree.JCParens;
import com.sun.tools.javac.tree.JCTree.JCSynchronized;
import com.sun.tools.javac.tree.JCTree.JCUnary;

public class ObjectRemover extends JmlTreeTranslator implements IJmlVisitor {

	ObjectSet objects;
	Stack<JCTree> removed;
	Stack<List<JCTree>> removedLists;

	public ObjectRemover(ObjectSet objects) {
		super();
		this.objects = objects;
		removed = new Stack<JCTree>();
		removedLists = new Stack<List<JCTree>>();
	}

	public <T extends JCTree> com.sun.tools.javac.util.List<T> translate(
			com.sun.tools.javac.util.List<T> trees) {
		if (trees == null)
			return null;
		boolean changed = false;
		List<JCTree> clone = new ArrayList<JCTree>();
		for (com.sun.tools.javac.util.List<T> l = trees; l.nonEmpty(); l = l.tail) {
			l.head = translate(l.head);
			if (l.head == null) {
				clone.add(removed.pop());
				changed = true;
			} else
				clone.add(l.head);
		}
		if (changed)
			removedLists.push(clone);
		return trees;
	}

	public void visitIf(JCIf tree) {
		JCIf clone = (JCIf) tree.clone();
		tree.cond = translate(tree.cond);
		tree.thenpart = translate(tree.thenpart);
		tree.elsepart = translate(tree.elsepart);
		if (tree.cond == null) {
			clone.cond = (JCExpression) removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitUnary(JCUnary tree) {
		JCUnary clone = (JCUnary) tree.clone();
		tree.arg = translate(tree.arg);
		if (tree.arg == null) {
			clone.arg = (JCExpression) removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitBinary(JCBinary tree) {
		JCBinary clone = (JCBinary) tree.clone();
		tree.lhs = translate(tree.lhs);
		tree.rhs = translate(tree.rhs);
		if (tree.lhs == null || tree.rhs == null) {
			if (tree.lhs == null && tree.rhs == null)
				removed.pop();
			removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitApply(JCMethodInvocation tree) {
		JCMethodInvocation clone = (JCMethodInvocation) tree.clone();
		tree.meth = translate(tree.meth);
		tree.args = translate(tree.args);
		if (tree.meth == null || tree.args == null || tree.args.contains(null)) {
			if (tree.args == null || tree.args.contains(null)) {
				com.sun.tools.javac.util.List<JCExpression> newclone = com.sun.tools.javac.util.List
						.nil();
				for (JCTree t : removedLists.pop()){
					newclone = newclone.append((JCExpression) t);
				}
				clone.args = newclone;
			}
			if (tree.meth == null)
				clone.meth = (JCExpression) removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitAssign(JCAssign tree) {
		JCAssign clone = (JCAssign) tree.clone();
		tree.lhs = translate(tree.lhs);
		tree.rhs = translate(tree.rhs);
		if (tree.lhs == null || tree.rhs == null) {
			if (tree.rhs == null)
				clone.rhs = (JCExpression) removed.pop();
			if (tree.lhs == null)
				clone.lhs = (JCExpression) removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitAssignop(JCAssignOp tree) {
		JCAssignOp clone = (JCAssignOp) tree.clone();
		tree.lhs = translate(tree.lhs);
		tree.rhs = translate(tree.rhs);
		if (tree.lhs == null || tree.rhs == null) {
			if (tree.rhs == null)
				clone.rhs = (JCExpression) removed.pop();
			if (tree.lhs == null)
				clone.lhs = (JCExpression) removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitSynchronized(JCSynchronized tree) {
		JCSynchronized clone = (JCSynchronized) tree.clone();
		tree.lock = translate(tree.lock);
		tree.body = translate(tree.body);
		if (tree.lock == null) {
			clone.lock = (JCExpression) removed.pop();
			removed.push(clone);
			result = tree.body;
		} else
			result = tree;
	}

	public void visitParens(JCParens tree) {
		JCParens clone = (JCParens) tree.clone();
		tree.expr = translate(tree.expr);
		if (tree.expr == null) {
			clone.expr = (JCExpression) removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitSelect(JCFieldAccess tree) {
		JCFieldAccess clone = (JCFieldAccess) tree.clone();
		tree.selected = translate(tree.selected);
		if (tree.selected == null) {
			clone.selected = (JCExpression) removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitIndexed(JCArrayAccess tree) {
		JCArrayAccess clone = (JCArrayAccess) tree.clone();
		tree.indexed = translate(tree.indexed);
		tree.index = translate(tree.index);
		if (tree.indexed == null || tree.index == null) {
			if (tree.index == null)
				clone.index = (JCExpression) removed.pop();
			if (tree.indexed == null)
				clone.indexed = (JCExpression) removed.pop();
			removed.push(clone);
			result = null;
		} else
			result = tree;
	}

	public void visitIdent(JCIdent tree) {
		if (objects.contains(tree.toString())) {
			removed.push(tree);
			result = null;
		} else
			result = tree;
	}

	public List<JCTree> getRemoved() {
		return removed;
	}

	public static List<JCTree> removeObjectStatements(JCTree tree, ObjectSet objects) {
		ObjectRemover translator = new ObjectRemover(objects);
		tree = translator.translate(tree);
		return translator.getRemoved();
	}

}
