package resanaplugin.costa;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Scanner;

import org.eclipse.ui.console.MessageConsoleStream;

import resanaplugin.ConsoleMessager;
import resanaplugin.Method;

public class Costa {
	private final boolean stack;
	private final String classPath;
	private final String classDeclaration;
	private Queue<Method> methods;
	
	private MessageConsoleStream console;

	public Costa(String classPath, String classDeclaration, Queue<Method> methods, boolean stack) throws IOException {
		this.classDeclaration = classDeclaration;
		this.classPath = classPath;
		this.methods = methods;
		this.stack = stack;
		console = ConsoleMessager.findConsole("Stack Analysis").newMessageStream();
	}
	
	private List<String> setExec(String method){
		
		List<String> exec = new ArrayList<String>();
		exec.add("costa");
		exec.add("-cp");
		exec.add(classPath);
		exec.add("-c");
		exec.add(classDeclaration);
		exec.add("-m");
		exec.add(method);
		if(stack){
			exec.add("-b");
			exec.add(classDeclaration);
			exec.add(method);
		}
		exec.add("cost_model");
		exec.add("=");
		if(stack)
			exec.add("ncalls");
		else
			exec.add("heap");
		exec.add("memory_unit");
		exec.add("=");
		exec.add("symbolic");
		exec.add("-v");
		exec.add("0");
		exec.add("go_into_java_api");
		exec.add("=");
		exec.add("true");
		//exec.add("interm_results");
		//exec.add("=");
		//exec.add("true");
		if(stack){
			exec.add("show_rf");
			exec.add("=");
			exec.add("yes");
		}
		exec.add("enable_field_sensitive");
		exec.add("=");
		exec.add("trackable");
		exec.add("enable_liveVariables");
		exec.add("=");
		exec.add("true");
		exec.add("enable_escape");
		exec.add("=");
		exec.add("true");
		exec.add("disj_neq_op");
		exec.add("=");
		exec.add("true");
		//exec.add("enable_slicing");
		//exec.add("=");
		//exec.add("true");
		exec.add("always_load_super_classes");
		exec.add("=");
		exec.add("yes");
		exec.add("static_initializers");
		exec.add("=");
		exec.add("true");
		exec.add("enable_sharing");
		exec.add("=");
		exec.add("topdown");
		return exec;
	}
	
	public String next() throws IOException {
			String methodSignature = methods.poll().getMethodSignature(false);
			console.println("Analysing method " + methodSignature + " with COSTA using the following command:");			
			List<String> exec = setExec(methodSignature);
			ConsoleMessager.printCollectionToConsole(console, exec, ' ');
			List<String> output = new LinkedList<String>();
			String[] execStr = new String[exec.size()];
			exec.toArray(execStr);
			Process p = Runtime.getRuntime().exec(execStr);
			Scanner scan = new Scanner(p.getInputStream());
			while (scan.hasNext())
				output.add(scan.nextLine());
			ConsoleMessager.printCollectionToConsole(console,output, '\n');
			return parse(output, classDeclaration + "_" + methodSignature);
	}

	public Method getNextMethod (){
		return methods.peek();
	}
	
	public boolean hasNext() {
		return methods.size() > 0;
	}
	
	public String parse(List<String> output, String methodSignature) {
		String result = null;
		for (int index = 0; index < output.size(); index++) {
			if (output.get(index).contains(
					"Upper Bound for '" + methodSignature + "'")) {
				result = output.get(index + 1);
				int lastSpace = result.lastIndexOf(' ');
				result = result.substring(0, lastSpace);
				if(!stack)
					break;
			}
			else if(stack && output.get(index).contains("Ranking function:")){
				result = output.get(index).replace("Ranking function: ", "");
				break;
			}
		}
		if(result == null)
			return "fail!";
		System.out.println(result);
		return result;
	}	
}
