package resanaplugin.costa;

import java.io.IOException;
import java.util.LinkedList;
import loopbounds.parsetree.MethodScanner;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.internal.ui.javaeditor.JavaEditor;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorActionDelegate;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.PlatformUI;
import org.jmlspecs.openjml.API;
import org.jmlspecs.openjml.JmlToken;
import org.jmlspecs.openjml.JmlTree;
import org.jmlspecs.openjml.JmlTree.JmlCompilationUnit;
import org.jmlspecs.openjml.JmlTree.JmlMethodClause;
import org.jmlspecs.openjml.JmlTree.JmlMethodClauseConditional;
import org.jmlspecs.openjml.JmlTree.JmlMethodDecl;
import org.jmlspecs.openjml.JmlTree.Maker;

import resanaplugin.ConsoleMessager;
import resanaplugin.Method;
import resanaplugin.ProjectInformation;

import com.sun.tools.javac.tree.JCTree.JCExpression;

@SuppressWarnings("restriction")
public class CostaEditorActionDelegate implements IEditorActionDelegate {
	private static final String COSTA_ID = "resanaplugin.costa.CostaEditorActionDelegate.costa";
	private ProjectInformation projectInf;
	private Maker maker;

	public CostaEditorActionDelegate() {
		projectInf = new ProjectInformation();
		try {
			API api = new API();
			api.parseString("temp", "class A{}");
			maker = JmlTree.Maker.instance(api.context());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public void run(IAction action) {
		if (action.getId().equals(COSTA_ID)) {
			Display display = Display.getCurrent();
			if (projectInf.save(display.getActiveShell())) {
				if (projectInf.errors())
					ConsoleMessager.compilationErrorDialog(display
							.getActiveShell());
				else {
					ICompilationUnit currentFile = projectInf.getCurrentFile();
					CostaOptions options = getCostaOptions(currentFile);

					CostaJob job = new CostaJob(display, options, currentFile);
					job.schedule();
				}
			}
		}
	}

	private class CostaJob extends Job {
		private final Display display;
		private final CostaOptions options;
		private final ICompilationUnit currentFile;

		public CostaJob(Display display, CostaOptions options,
				ICompilationUnit currentFile) {
			super("Running Costa and JaSB");
			this.display = display;
			this.options = options;
			this.currentFile = currentFile;
		}

		@Override
		public IStatus run(IProgressMonitor monitor) {
			String classDeclaration = ProjectInformation
					.getClassDeclaration(currentFile);
			Costa costa;
			IJavaProject project = projectInf.getProject();
			String classpath = ProjectInformation.getClassPath(project);
			try {
				costa = new Costa(classpath, classDeclaration,
						new LinkedList<Method>(options.methods),
						options.measuredBy);
			} catch (IOException e) {
				e.printStackTrace();
				throw new IllegalStateException(e);
			}
			monitor.beginTask("Running Costa", options.methods.size());
			while (costa.hasNext()) {
				final Method method = costa.getNextMethod();
				String result;
				try {
					result = costa.next();
				} catch (Exception e) {
					result = "fail!";
				}
				if (options.measuredBy) {
					JaSB jasb = new JaSB(result, classpath,false,
							options.vm,true);
					final JCExpression measureBy = jasb.convertToJML(maker);
					display.syncExec(new Runnable() {

						@Override
						public void run() {
							IRegion region = projectInf.getViewer()
									.getRangeIndication();
							replaceMeasuredBy(method.getMethod(), measureBy,
									null);
							projectInf.getViewer().setRangeIndication(
									region.getOffset(), region.getLength(),
									true);
						}
					});
				} else {
					final String[] bounds = new String[3];
					if (options.costaBound)
						bounds[0] = result;
					if (options.simpleBound) {
						JaSB jasb = new JaSB(result, classpath, false, options.vm, false);
						bounds[1] = jasb.start();
					}
					if (options.concreteBound) {
						JaSB jasb = new JaSB(result, classpath, true, options.vm, false);
						bounds[2] = jasb.start();
					}
					display.syncExec(new Runnable() {

						@Override
						public void run() {
							IRegion region = projectInf.getViewer()
									.getRangeIndication();
							if (options.costaBound)
								replaceComment(method.getMethod(),
										"costaHeapBound", bounds[0]);
							if (options.simpleBound)
								replaceComment(method.getMethod(),
										"simpleHeapBound", bounds[1]);
							if (options.concreteBound)
								replaceComment(method.getMethod(),
										"concreteHeapBound", bounds[2]);
							projectInf.getViewer().setRangeIndication(
									region.getOffset(), region.getLength(),
									true);
						}
					});
				}
				monitor.worked(1);
			}
			monitor.done();
			return Status.OK_STATUS;
		}
	}

	private CostaOptions getCostaOptions(ICompilationUnit currentFile) {
		Shell sh = PlatformUI.getWorkbench().getActiveWorkbenchWindow()
				.getShell();
		CostaOptionsDialog costaDialog = new CostaOptionsDialog(sh);
		CostaOptions i = costaDialog.createWindow(ProjectInformation
				.getMethods(currentFile));
		return i;
	}

	@Override
	public void selectionChanged(IAction action, ISelection selection) {
		projectInf.changeSelection(selection);
	}

	@Override
	public void setActiveEditor(IAction action, IEditorPart targetEditor) {
		if (targetEditor instanceof JavaEditor) {
			projectInf.setActiveEditor((JavaEditor) targetEditor);
		}
	}

	private boolean replaceComment(IMethod method, String name, String value) {
		try {
			String seperator = System.getProperty("line.separator");
			String source = projectInf.getViewer().getDocument().get();
			int methodPos = method.getNameRange().getOffset();
			int lineBeforeMethod = source.lastIndexOf("\n", methodPos);
			String[] lines = source.substring(0, lineBeforeMethod).split("\n");
			String ident = source.substring(lineBeforeMethod + 1, methodPos);
			ident = ident.substring(0, ident.length() - ident.trim().length()
					- 1);
			boolean done = false;
			for (int i = lines.length - 1; i >= 0; i--) {
				String line = lines[i].trim();

				if (!line.startsWith("//#")) {
					break;
				}
				line = line.substring(3).trim();
				if (line.startsWith(name)) {
					line = ident + "//# " + name + " " + value;
					done = true;
					lines[i] = line;
				}
			}
			String result = "";
			for (String line : lines) {
				result += line + seperator;
			}
			if (!done) {
				result += ident + "//# " + name + " " + value + seperator;
			}
			source = result + source.substring(lineBeforeMethod + 1);
			IDocument document = projectInf.getViewer().getDocument();
			document.set(source);
			int newMethodPos = method.getNameRange().getOffset();
			// Wait for eclipse to update the IMethod
			int i = 1000;
			while (i-- > 0 && newMethodPos == methodPos) {
				Thread.sleep(1);
				newMethodPos = method.getNameRange().getOffset();
			}
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}

	private boolean replaceMeasuredBy(IMethod method, JCExpression expression,
			JCExpression predicate) {
		try {
			String source = projectInf.getViewer().getDocument().get();
			if (method == null) {
				System.err.println("Method is null");
				return false;
			}

			int methodPos = method.getNameRange().getOffset();

			API api = new API();

			JmlCompilationUnit compUnit = api.parseString("jml", source);
			JmlMethodDecl jmlMethod = null;
			for (JmlMethodDecl tempMethod : MethodScanner.getMethods(compUnit)) {
				if (tempMethod.pos == methodPos) {
					jmlMethod = tempMethod;
				}
			}
			if (jmlMethod == null) {
				System.err.println("Method not found in JML parser");
				return false;
			}
			int specStart = -1;
			int specEnd = -1;
			if (jmlMethod.cases != null) {
				specStart = source.lastIndexOf("/*@", jmlMethod.cases.pos);
				if (specStart != -1) {
					specEnd = source.indexOf("*/", specStart) + 2;
				}
			}
			JmlMethodClauseConditional measuredByClause = MeasureByScanner
					.findMeasuredBy(jmlMethod);
			if (measuredByClause == null) {
				Maker maker = JmlTree.Maker.instance(api.context());
				measuredByClause = maker.JmlMethodClauseConditional(
						JmlToken.MEASURED_BY, expression, predicate);
				if (jmlMethod.cases == null) {
					jmlMethod.cases = maker.JmlMethodSpecs(null);
				}
				if (jmlMethod.cases.cases == null) {
					jmlMethod.cases.cases = com.sun.tools.javac.util.List
							.of(null);
				}
				if (jmlMethod.cases.cases.head == null) {
					com.sun.tools.javac.util.List<JmlMethodClause> list = com.sun.tools.javac.util.List
							.of(null);
					jmlMethod.cases.cases.head = maker.JmlSpecificationCase(
							maker.Modifiers(0), false, null, null, list);
				}
				if (jmlMethod.cases.cases.head.clauses == null) {
					jmlMethod.cases.cases.head.clauses = com.sun.tools.javac.util.List
							.of(null);
				}
				if (jmlMethod.cases.cases.head.clauses.head == null) {
					jmlMethod.cases.cases.head.clauses.head = measuredByClause;
				} else {
					com.sun.tools.javac.util.List<JmlMethodClause> otherClauses = jmlMethod.cases.cases.head.clauses;
					jmlMethod.cases.cases.head.clauses = com.sun.tools.javac.util.List
							.of((JmlMethodClause) measuredByClause);
					jmlMethod.cases.cases.head.clauses.setTail(otherClauses);
				}
			} else {
				measuredByClause.expression = expression;
				measuredByClause.predicate = predicate;
			}
			int lineBeforeMethod = source.lastIndexOf("\n", jmlMethod.pos);
			if (specStart == -1) {
				specStart = lineBeforeMethod;
				specEnd = lineBeforeMethod;
			}
			String ident = source
					.substring(lineBeforeMethod + 1, jmlMethod.pos);
			ident = ident.substring(0, ident.length() - ident.trim().length()
					- 1);
			String spec = jmlMethod.cases.toString();
			spec = spec.replace("\n", "\n" + ident);
			String newSource = source.substring(0, specStart) + spec
					+ source.substring(specEnd, source.length());
			projectInf.getViewer().getDocument().set(newSource);
			int newMethodPos = method.getNameRange().getOffset();
			// Wait for eclipse to update the IMethod
			int i = 1000;
			while (i-- > 0 && newMethodPos == methodPos) {
				Thread.sleep(1);
				newMethodPos = method.getNameRange().getOffset();
			}
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}
}
