package resanaplugin.costa;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.Vector;

import org.jmlspecs.openjml.API;
import org.jmlspecs.openjml.JmlTree.Maker;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCExpression;
import com.sun.tools.javac.tree.JCTree.JCFieldAccess;
import com.sun.tools.javac.util.List;

//import javax.realtime.RealtimeThread;
//import javax.realtime.VTMemory;

/* grammar
 * expr -> addexpr
 * addexpr -> mulexpr (('+'/'-') mulexpr)*
 * mulexpr -> primary (('*'/'%') primary)*
 * primary -> '(' expr ')' | number | 'c' (' 'size' '(' bla ')' ')' | 'nat' '(' text ')'
 */

/*
class MemoryTest extends RealtimeThread{
	public void run(){
		final VTMemory sma = new VTMemory(32*1024);
		sma.enter(new Runnable(){
			public void run(){
				//long before_alt = Runtime.getRuntime().freeMemory();
				Constructor cls = null;
				try {
					cls = Class.forName("java.lang.Integer").getConstructor();
				} catch (Exception e) {
				}
				long before = getCurrentMemoryArea().memoryConsumed();
				try {
					//Object a = (Object)cls.newInstance();
					Integer a = new Integer(0);
				} catch (Exception e) {
				}
				long after=getCurrentMemoryArea().memoryConsumed();
				System.out.println(after-before);
				//System.out.println(i + ": bytearray=" + byteArraySize + " chararray=" + charArraySize + " string=" + stringSize);
				//System.out.println("change: " + (after-before) + " alt=" + (after_alt-before_alt));
			}
		});
	}
}
*/
enum TokenType {
    TEXT, NUMBER, PLUS, MIN, MUL, OPEN, CLOSE, CONSTANT, SIZE, VARIABLE, COMMA, MAXIMIZE_FAILED, PRIMITIVE_TYPE, REFTYPE, EOF, POW, MAX, OPEN_SQUARE, CLOSE_SQUARE
}

class Token<E> {
    TokenType type;
    E token;
    String text;
    Token(TokenType type, String text) {
        this.type = type;
        this.token = null;
        this.text = text;
    }
    Token(TokenType type, String text, E token) {
        this.type = type;
        this.token = token;
        this.text = text;
    }

    public String toString() {
        if (token != null)
            return type + ": " + token.toString();
        return type.toString();
    }
}

abstract class Value {
    int l;
    Value(int l) {
        this.l = l;
    }
    public abstract boolean isConcrete();
    public String toString(int l) {
        String str = toString();
        if (level() >= l) //|| level() == PRIMARY_LEVEL)
            return str;
        return "(" + str + ")";
    }
    public abstract String toString();

    final static int PRIMARY_LEVEL = 4;
    final static int POW_LEVEL = 3;
    final static int MUL_LEVEL = 2;
    final static int PLUS_LEVEL = 1;
    final static int TOP_LEVEL = 0;

    public int level() {
        return l;
    }

    public abstract	boolean hasTerm(Value v);
    public abstract Value countTerm(Value v, boolean forceTotal);
    public Value timesTerm(Value v) {
        return countTerm(v, true);
    }
    // return all numbers and variables
    public abstract void allTerms(Vector<Value> terms);

    public abstract Value without(Value a);
	public abstract JCExpression translateToJML(Maker maker);
}

class Number extends Value {
    int number;
    Number(int number) {
        super(PRIMARY_LEVEL);
        this.number = number;
    }
    public String toString() {
        return new Integer(number).toString();
    }

    public boolean isConcrete() {
        return true;
    }
    public int intValue() {
        return number;
    }

    public boolean equals(Object b) {
        return b instanceof Number && ((Number)b).intValue() == intValue();
    }

    public boolean hasTerm(Value v) {
        if (v instanceof Number) {
            Number n = (Number)v;
            return n.intValue() == intValue();
        }
        return false;
    }

    public void allTerms(Vector<Value> terms) {
        terms.add(this);
    }

    public Value countTerm(Value v, boolean forceTotal) {
        return hasTerm(v) ? new Number(1) : null;
    }

    public Value without(Value x) {
        if (this.equals(x))
            return null;
        return this;
    }
    
	@Override
	public JCExpression translateToJML(Maker maker) {
		return maker.Literal(number);
	}
}

class Negation extends Value {
    Value value;
    Negation(Value value) {
        super(PRIMARY_LEVEL);
        this.value = value;
    }

    public String toString() {
        return "-" + value.toString(PRIMARY_LEVEL);
    }

    public boolean isConcrete() {
        return value.isConcrete();
    }

    public boolean equals(Object b) {
        return b instanceof Negation && ((Negation)b).value.equals(value);
    }

    public boolean hasTerm(Value v) {
        return value.hasTerm(v);
    }

    public void allTerms(Vector<Value> terms) {
        value.allTerms(terms);
    }

    public Value countTerm(Value v, boolean forceTotal) {
        Value count = value.countTerm(v, forceTotal);
        return count != null ? new Negation(count) : null;
    }

    public Value without(Value x) {
        Value wo = value.without(x);
        return wo != null ? new Negation(wo) : null;
    }
    
	@Override
	public JCExpression translateToJML(Maker maker) {
		return maker.Unary(JCTree.NEG, value.translateToJML(maker));
	}
}

class Variable extends Value {
    String name;

    Variable(String name) {
        super(PRIMARY_LEVEL);
        this.name = name;
    }
    public String toString() {
        return name;
    }

    public boolean isConcrete() {
        return false;
    }

    public boolean equals(Object b) {
        return b instanceof Variable && ((Variable)b).toString().equals(toString());
    }

    public boolean hasTerm(Value v) {
        if (v instanceof Variable) {
            Variable var = (Variable)v;
            return name.equals(var.name);
        }
        return false;
    }

    public void allTerms(Vector<Value> terms) {
        terms.add(this);
    }

    public Value countTerm(Value v, boolean forceTotal) {
        return hasTerm(v) ? new Number(1) : null;
    }

    public Value without(Value x) {
        if (this.equals(x))
            return null;
        return this;
    }
    
	@Override
	public JCExpression translateToJML(Maker maker) {
		return maker.Ident(name);
	}
}

abstract class BinaryOperation extends Value {
    Value a;
    Value b;
    BinaryOperation(Value a, Value b, int level) {
        super(level);
        this.a = a;
        this.b = b;
    }

    public boolean isConcrete() {
        return a.isConcrete() && b.isConcrete();
    }

    public abstract boolean isSymmetrical();

    public Value firstComponent() {
        return a;
    }
    public Value secondComponent() {
        return b;
    }

    public void setFirstComponent(Value a) {
        this.a = a;
    }
    public void setSecondComponent(Value b) {
        this.b = b;
    }
    public boolean equals(Object _alt) {
        if (!(_alt instanceof BinaryOperation))
            return false;
        BinaryOperation alt = (BinaryOperation)_alt;
        return getClass().equals(alt.getClass())
               && a.equals(alt.firstComponent())
               && b.equals(alt.secondComponent());
    }
    public abstract Value create(Value a, Value b);
}

class AdditionOperation extends BinaryOperation {
    AdditionOperation(Value a, Value b) {
        super(a, b, PLUS_LEVEL);
    }
    public String toString() {
        if (b instanceof Negation) {
            Negation n = (Negation)b;
            return a.toString(level()) + "-" + n.value.toString(level()+1);
        } else if (b instanceof Number && ((Number)b).intValue() < 0) {
            return a.toString(level()) + "-" + new Integer(-((Number)b).intValue()).toString();
        }
        return a.toString(level()) + "+" + b.toString(level());
    }

    public boolean isSymmetrical() {
        return true;
    }

    public AdditionOperation create(Value a, Value b) {
        return new AdditionOperation(a,b);
    }
    public boolean hasTerm(Value v) {
        return a.hasTerm(v) || b.hasTerm(v);
    }
    public void allTerms(Vector<Value> terms) {
        a.allTerms(terms);
        b.allTerms(terms);
        terms.add(this);
    }
    public Value countTerm(Value v, boolean forceTotal) {
        if (this.equals(v)) {
            return new Number(1);
        }
        Value ca = a.countTerm(v, forceTotal);
        Value cb = b.countTerm(v, forceTotal);
        if (ca != null && cb != null)
            return new AdditionOperation(ca, cb);
        if (forceTotal)
            return null;
        return ca != null ? ca : cb;
    }
    public Value without(Value x) {
        if (this.equals(x))
            return null;
        if (a.equals(x))
            return b;
        if (b.equals(x))
            return a;
        /*
        Value woa = a.without(x);
        if (woa.equals(a)) {
        	// nothing removed
        	return new AdditionOperation(a,b.without(x));
        }
        */
        return new AdditionOperation(a.without(x), b.without(x));
    }
    
	@Override
	public JCExpression translateToJML(Maker maker) {
		JCExpression lhs = a.translateToJML(maker);
		JCExpression rhs = b.translateToJML(maker);
		return maker.Binary(JCTree.PLUS, lhs, rhs);
	}
}

class MultiplicationOperation extends BinaryOperation {
    MultiplicationOperation(Value a, Value b) {
        super(a, b, MUL_LEVEL);
    }
    public String toString() {
        return a.toString(level()) + "*" + b.toString(level());
    }

    public boolean isSymmetrical() {
        return true;
    }
    public MultiplicationOperation create(Value a, Value b) {
        return new MultiplicationOperation(a,b);
    }
    public boolean hasTerm(Value v) {
        return a.equals(v) || b.equals(v);
    }
    public void allTerms(Vector<Value> terms) {
        a.allTerms(terms);
        b.allTerms(terms);
        terms.add(this);
    }
    public Value countTerm(Value v, boolean forceTotal) {
        if (this.equals(v)) {
            return new Number(1);
        }
        Value ca = a.countTerm(v, forceTotal);
        Value cb = b.countTerm(v, forceTotal);
        if (ca != null)
            return new MultiplicationOperation(ca, b);
        if (cb != null)
            return new MultiplicationOperation(a, cb);
        return null;
    }
    public Value without(Value x) {
        if (this.equals(x))
            return null;
        /*
        if (a.equals(x))
        	return null;
        if (b.equals(x))
        	return null;
        */
        /*
        Value woa = a.without(x);
        Value wob = b.without(x);
        if (woa == null)
        	return wob;
        if (wob == null)
        	return woa;
        return new MultiplicationOperation(woa, wob);
        /*
        if (a.equals(x))
        	return b;
        if (b.equals(x))
        	return a;
        */
        Value woa = a.without(x);
        if (woa == null)
            return null;
        // something is removed
        if (!woa.equals(a))
            return new MultiplicationOperation(woa, b);
        Value wob = b.without(x);
        if (wob == null)
            return null;
        return new MultiplicationOperation(a, wob);
    }
    
	@Override
	public JCExpression translateToJML(Maker maker) {
		JCExpression lhs = a.translateToJML(maker);
		JCExpression rhs = b.translateToJML(maker);
		return maker.Binary(JCTree.MUL, lhs, rhs);
	}
}

class PowerOperation extends BinaryOperation {
    PowerOperation(Value a, Value b) {
        super(a, b, POW_LEVEL);
    }
    public String toString() {
        return a.toString(level()) + "^" + b.toString(level());
    }
    public boolean isSymmetrical() {
        return false;
    }
    public PowerOperation create(Value a, Value b) {
        return new PowerOperation(a,b);
    }
    public boolean hasTerm(Value v) {
        return a.equals(v);
    }
    public void allTerms(Vector<Value> terms) {
        terms.add(this);
        a.allTerms(terms);
    }
    public Value countTerm(Value v, boolean forceTotal) {
        if (this.equals(v)) {
            return new Number(1);
        }
        if (b.isConcrete() && a.equals(v)) // && !forceTotal)
            return new PowerOperation(a, new AdditionOperation(b, new Number(-1)));
        //if (a.equals(v))
        //        return this;
        return null;
    }
    public Value timesTerm(Value v) {
        if (this.equals(v)) {
            return new Number(1);
        }
        if (a.equals(v))
            return b;
        return null;
    }
    public Value without(Value x) {
        if (this.equals(x))
            return null;
        if (a.equals(x)) {
            return null;
        }
        return this;
    }
    
	@Override
	public JCExpression translateToJML(Maker maker) {
		JCExpression lhs = a.translateToJML(maker);
		JCExpression rhs = b.translateToJML(maker);
		JCFieldAccess meth = maker.Select(maker.Ident("Math"),
				maker.Name("pow"));
		return maker.Apply(null, meth, List.of(lhs, rhs));
	}
}

class MaxOperation extends BinaryOperation{

	MaxOperation(Value a, Value b) {
		super(a, b, PRIMARY_LEVEL);
	}

	@Override
	public boolean isSymmetrical() {
		return true;
	}

	@Override
	public Value create(Value a, Value b) {
		return new MaxOperation(a, b);
	}

	@Override
	public String toString() {
		return "max(" + a.toString(level()) + "," + b.toString(level()) + ")";
	}

	@Override
	public boolean hasTerm(Value v) {
		return a.equals(v) ? true : b.equals(v);
	}

	@Override
	public Value countTerm(Value v, boolean forceTotal) {
		if (this.equals(v)) {
            return new Number(1);
        }
        Value ca = a.countTerm(v, forceTotal);
        Value cb = b.countTerm(v, forceTotal);
        if (ca != null)
            return new MaxOperation(ca, b);
        if (cb != null)
            return new MaxOperation(a, cb);
        return null;
	}

	@Override
	public void allTerms(Vector<Value> terms) {
		terms.add(this);
		a.allTerms(terms);
		b.allTerms(terms);
	}

	@Override
	public Value without(Value x) {
		Value aa = a.without(x);
		Value bb = b.without(x);
		if(aa == null || bb == null)
			return null;
		return new MaxOperation(aa, bb);
	}

	@Override
	public JCExpression translateToJML(Maker maker) {
		JCExpression lhs = a.translateToJML(maker);
		JCExpression rhs = b.translateToJML(maker);
		JCFieldAccess meth = maker.Select(maker.Ident("Math"),
				maker.Name("max"));
		return maker.Apply(null, meth, List.of(lhs, rhs));
	}
}

/**
 * Recursive_descent_parser
 * @author bvgastel
 */
public class JaSB {
    StringTokenizer s;
    int pos;
    StringBuffer before;
    String classPath;
    boolean calculateSize;
    String vm;
    boolean removeNativeCalls;
    TreeMap<String, Integer> sizeCache;

    private static int runJamaicaProcess(String command) throws Exception {
    	String envp[] = new String[1];
    	envp[0] = "JAMAICAVM_SCOPEDSIZE=32M";
        System.out.println("Going to execute jamaica");
        Process pro = Runtime.getRuntime().exec(command, envp);
        //printLines(command + " stdout:", pro.getInputStream());
        //printLines(command + " stderr:", pro.getErrorStream());
        System.out.println("Executing jamaica");
        pro.waitFor();
        System.out.println(command + ": " + pro.exitValue());
        return pro.exitValue();
    }
    
    private static int runHostProcess(String command) throws Exception {
        System.out.println("Going to execute javac");
        Process pro = Runtime.getRuntime().exec(command);
        //printLines(command + " stdout:", pro.getInputStream());
        //printLines(command + " stderr:", pro.getErrorStream());
        System.out.println("Executing javac");
        pro.waitFor();
        System.out.println(command + ": " + pro.exitValue());
        return pro.exitValue();
    }

    private static int doJamaicaTestSize(String object, String classPath) throws Exception {
    	object = object.replace("$", ".");
        try {
            // Create file
            FileWriter fstream = new FileWriter("memorytest.java");
            BufferedWriter out = new BufferedWriter(fstream);
            out.write("import javax.realtime.RealtimeThread;import javax.realtime.VTMemory;public class memorytest extends RealtimeThread {public void run(){final VTMemory sma = new VTMemory(32*1024);sma.enter(new Runnable(){public void run(){long before =getCurrentMemoryArea().memoryConsumed();" +
                      object + " a = new " + object + "();" +
                      "long after=getCurrentMemoryArea().memoryConsumed();int size = (int)(after-before);System.exit(size);}});}public static void main(String[] args) {new memorytest().start();}}");

            //Close the output stream
            out.close();
        } catch (Exception e) { //Catch exception if any
            System.err.println("Error: " + e.getMessage());
        }
        System.err.println("Calculating size of " + object);
        if (runJamaicaProcess("jamaicac -classpath " + classPath + " memorytest.java") != 0) {
            throw new Exception("calculating " + object + " failed");
        }
        int size = runJamaicaProcess("jamaicavm -classpath .:" + classPath + " memorytest");
        System.err.println(" = " + size);
        return size;
    }
    
    private static int doHostTestSize(String object, String classPath) throws Exception {
    	object = object.replace("$", ".");
        try {
            // Create file
            FileWriter fstream = new FileWriter("memorytestopen.java");
            BufferedWriter out = new BufferedWriter(fstream);
			out.write("import java.lang.management.*; public class memorytestopen {static void collectGarbage() { for (int i = 0; i < 5; i++) { try { System.gc(); Thread.currentThread().sleep(10); System.runFinalization(); Thread.currentThread().sleep(10); } catch (InterruptedException ex) { ex.printStackTrace(); } } } public static void main(String args[]) { Runtime.getRuntime().totalMemory(); Runtime.getRuntime().freeMemory(); collectGarbage(); long before = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory(); collectGarbage(); "
					+ object + " a = new " + object + "(); collectGarbage(); long after = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory(); int size = (int)(after-before); Object b = a; System.exit(size); } }");
            //Close the output stream
            out.close();
        } catch (Exception e) { //Catch exception if any
            System.err.println("Error: " + e.getMessage());
        }
        System.err.println("Calculating size of " + object);
        if (runHostProcess("javac -classpath " + classPath + " memorytestopen.java") != 0) {
            throw new Exception("calculating " + object + " failed");
        }
        int size = runHostProcess("java -classpath .:" + classPath + " memorytestopen");
        System.err.println(" = " + size);
        return size;
    }
    
    private int testSize(String object, String classPath) throws Exception {
    	Integer result = sizeCache.get(object);
    	if (result == null) {
    		int size;
    		if(vm == "jamaica")
    			size = doJamaicaTestSize(object, classPath);
    		else
    			size = doHostTestSize(object, classPath);
    		result = new Integer(size);
    		sizeCache.put(object, result);
    	}
    	return result.intValue();
    }
    
    JaSB(String expr, String classPath, boolean calculateSize, String vm, boolean removeNativeCalls) {
        s = new StringTokenizer(
            expr, //"nat(n)*c(size(MyObject, 1))",
            //"(nat(n)*3)*(nat(b)*2)",
            //"(1+2)*3+c(size(java/lang/String,1))+nat(n+1+2)+nat(b)+2",
            //"(nat(n)*3)+nat(b)*2*4",
            //"nat(n+1)* (c(size(java/lang/StringBuilder,1))+c(valueOf(Object))+c(<init>(String))+c(append(String))+c(toString()))+c(size(java/lang/StringBuilder,1))+c(valueOf(Object))+c(<init>(String))",
            //"c(size(Test$Container,1))+nat(i)*c(maximize_failed)*c(size(primitiveType(int),2))",
            //"nat(heap/AllocateThread.referenceObjectCount)*c(size(java/lang/Object,1))",
            //"nat(heap/AllocateThread.referenceObjectCount)*c(size(refType(java/lang/Object),1))+c(nativeArrayCopy(ObjectIObjectII))+c(nativeCurrentTimeMillis())",
            //"nat(bytes)*c(size(primitiveType(byte),1))",
            //"(pow(2,nat(1*n+ -1))-1)* (c(size(java/lang/StringBuilder,1))+c(valueOf(Object))+c(<init>(String))+c(append(String))+c(toString()))",
            "+-*()[],", true); // System.in);
        before = new StringBuffer();
        this.classPath = classPath;
        this.calculateSize = calculateSize;
        this.vm = vm;
        this.sizeCache = new TreeMap<String, Integer>();
        this.removeNativeCalls = removeNativeCalls;
    }

    void error(String str) throws Exception {
        System.out.print(before);
        if (current.text != null)
            System.out.print(current.text);
        while (s.hasMoreTokens())
            System.out.print(s.nextToken());
        System.out.println();
        for (int i = 0; i < pos; i++)
            System.out.print(" ");
        if (current.text == null)
            current.text = " ";
        for (int i = 0; i < current.text.length(); i++)
            System.out.print("^");
        System.out.println();
        throw new Exception(str + " (pos: " + pos + ")");
    }
    boolean hasNext() {
        return s.hasMoreTokens();
    }

    Token<?> current;
    Token<?> nextToken;

    Token<?> next() {
        String str;
        do {
            if (!s.hasMoreTokens())
                return new Token<Object>(TokenType.EOF, null);
            str = s.nextToken();
            str = str.trim();
        } while (str.length() == 0);
        if ("pow".equals(str))
            return new Token<Object>(TokenType.POW, str);
        if ("*".equals(str))
            return new Token<Object>(TokenType.MUL, str);
        if ("+".equals(str))
            return new Token<Object>(TokenType.PLUS, str);
        if ("(".equals(str))
            return new Token<Object>(TokenType.OPEN, str);
        if (")".equals(str))
            return new Token<Object>(TokenType.CLOSE, str);
        if ("nat".equals(str))
            return new Token<Object>(TokenType.VARIABLE, str);
        if ("c".equals(str))
            return new Token<Object>(TokenType.CONSTANT, str);
        if ("size".equals(str))
            return new Token<Object>(TokenType.SIZE, str);
        if (",".equals(str))
            return new Token<Object>(TokenType.COMMA, str);
        if ("refType".equals(str))
            return new Token<Object>(TokenType.REFTYPE, str);
        //if ("maximize_failed".equals(str))
        //	return new Token<Object>(TokenType.MAXIMIZE_FAILED, str);
        if ("primitiveType".equals(str))
            return new Token<Object>(TokenType.PRIMITIVE_TYPE,str);
        if ("max".equals(str))
        	return new Token<Object>(TokenType.MAX,str);
        if ("[".equals(str))
        	return new Token<Object>(TokenType.OPEN_SQUARE,str);
        if ("]".equals(str))
        	return new Token<Object>(TokenType.CLOSE_SQUARE,str);
        try {
            Integer i = new Integer(str);
            return new Token<Integer>(TokenType.NUMBER, str, i);
        } catch (Exception e) {
        }
        if ("-".equals(str))
            return new Token<Object>(TokenType.MIN, str);
        return new Token<String>(TokenType.TEXT, str, str);
    }

    boolean lookahead(TokenType type) {
        return current.type == type;
    }

    Token<?> get() {
        Token<?> retval = current;
        if (current != null) {
            pos += current.text.length();
            before.append(current.text);
        }
        current = nextToken != null ? nextToken : next();
        nextToken = null;
        /*
        if (lookahead(TokenType.MIN)) {
            nextToken = next();
            if (nextToken.type == TokenType.NUMBER) {
                Token<Integer> n = (Token<Integer>)nextToken;
                n.token = -n.token;
                n.text = current.text + n.text;
                current = n;
                nextToken = null;
            }
        }
        */
        //System.out.println("Token: " + current.toString());
        return retval;
    }

    boolean lookaheadPrimary() {
        return lookahead(TokenType.NUMBER)
               || lookahead(TokenType.OPEN)
               || lookahead(TokenType.CONSTANT)
               || lookahead(TokenType.VARIABLE)
               || lookahead(TokenType.POW)
               || lookahead(TokenType.TEXT)
               || lookahead(TokenType.MIN)
               || lookahead(TokenType.MAX);
    }
    Value primary() throws Exception {
        if (lookahead(TokenType.TEXT)) {
            return new Variable(get().text);
        }
        if (lookahead(TokenType.OPEN)) {
            get();
            Value retval = expr();
            if (!lookahead(TokenType.CLOSE))
                error("expected ')'");
            get();
            return retval;
        }
        if (lookahead(TokenType.POW)) {
            get();
            Value retval = null;
            if (!lookahead(TokenType.OPEN))
                error("expected '('");
            get();
            if (!lookaheadPrimary())
                error("expected primary()");
            Value a = primary();
            if (!lookahead(TokenType.COMMA))
                error("expected ','");
            get();
            if (!lookaheadPrimary())
                error("expected primary()");
            Value b = primary();
            retval = new PowerOperation(a, b);
            if (!lookahead(TokenType.CLOSE))
                error("expected ')'");
            get();
            return retval;
        }
        if (lookahead(TokenType.MAX)){
        	get();
        	Value retval = null;
        	if (!lookahead(TokenType.OPEN))
                error("expected '('");
        	get();
        	if (!lookahead(TokenType.OPEN_SQUARE))
        		error("expected '['");
        	get();
        	if (!lookaheadPrimary())
        		error("expected expr()");
        	Value a = expr();
        	if(!lookahead(TokenType.COMMA))
        		error("expected ','");
        	get();
        	if (!lookaheadPrimary())
        		error("expected expr()");
        	Value b = expr();
        	retval = new MaxOperation(a, b);
        	if(!lookahead(TokenType.CLOSE_SQUARE))
        		error("expected ']'");
        	get();
        	if(!lookahead(TokenType.CLOSE))
        		error("expected ')'");
        	get();
        	return retval;
        }
        if (lookahead(TokenType.CONSTANT)) {
            Value retval = null;
            get();
            if (!lookahead(TokenType.OPEN))
                error("expected '('");
            get();
            if (!lookahead(TokenType.MAXIMIZE_FAILED) && !lookahead(TokenType.SIZE) && !lookahead(TokenType.TEXT))
                error("expected 'size' or 'maximize_failed' or constant name");
            if (lookahead(TokenType.TEXT)) {
                Token<?> text = get();
                String out = text.text;
                boolean function = false;
                if (lookahead(TokenType.OPEN)) {
                	function = true;
                    text = get();
                    out += text.text;
                    while (!lookahead(TokenType.CLOSE)) {
                        text = get();
                        out += text.text;
                    }
                    text = get();
                    out += text.text;
                }
                if(removeNativeCalls && function)
                	retval = new Number(0);
                else
                	retval = new Variable(out);
            } else if (lookahead(TokenType.SIZE)) {
            	if (calculateSize) {
	                int retSize = 0;
	                get();
	                if (!lookahead(TokenType.OPEN))
	                    error("expected '('");
	                get();
	                if (!lookahead(TokenType.PRIMITIVE_TYPE) && !lookahead(TokenType.TEXT) && !lookahead(TokenType.REFTYPE))
	                    error("expected class name or 'primitiveType'");
	                if (lookahead(TokenType.REFTYPE)) {
	                    get();
	                    if (!lookahead(TokenType.OPEN))
	                        error("expected '('");
	                    get();
	                    if (!lookahead(TokenType.TEXT))
	                        error("expected text");
	                    get();
	                    if (!lookahead(TokenType.CLOSE))
	                        error("expected ')'");
	                    get();
	                    retSize = 4;
	                } else if (lookahead(TokenType.TEXT)) {
	                    Token<?> text = get();
	                    retSize = testSize(text.text.replace('/', '.'), classPath);
	                } else {
	                    // primitive type
	                    get();
	                    if (!lookahead(TokenType.OPEN))
	                        error("expected '('");
	                    get();
	                    if (!lookahead(TokenType.TEXT))
	                        error("expected text");
	                    Token<?> text = get();
	                    if (!lookahead(TokenType.CLOSE))
	                        error("expected ')'");
	                    get();
	                    if ("int".equals(text.token)) {
	                        retSize = 4;
	                    } else if ("byte".equals(text.token)) {
	                        retSize = 1;
	                    } else if ("short".equals(text.token)) {
	                        retSize = 2;
	                    } else if ("char".equals(text.token)) {
	                        retSize = 2;
	                    } else if ("long".equals(text.token)) {
	                        retSize = 8;
	                    } else if ("float".equals(text.token)) {
	                        retSize = 4;
	                    } else if ("double".equals(text.token)) {
	                        retSize = 8;
	                    }
	                }
	                if (!lookahead(TokenType.COMMA))
	                    error("expected ','");
	                get();
	                if (!lookahead(TokenType.NUMBER))
	                    error("expected integer");
	                get();
	                if (!lookahead(TokenType.CLOSE))
	                    error("expected ')'");
	                get();
	                retval = new Number(retSize);
            	} else {
	                get();
	                if (!lookahead(TokenType.OPEN))
	                    error("expected '('");
	                get();
	                if (!lookahead(TokenType.PRIMITIVE_TYPE) && !lookahead(TokenType.TEXT) && !lookahead(TokenType.REFTYPE))
	                    error("expected class name or 'primitiveType'");
	                if (lookahead(TokenType.REFTYPE)) {
	                    get();
	                    if (!lookahead(TokenType.OPEN))
	                        error("expected '('");
	                    get();
	                    if (!lookahead(TokenType.TEXT))
	                        error("expected text");
	                    get();
	                    if (!lookahead(TokenType.CLOSE))
	                        error("expected ')'");
	                    get();
	                    retval = new Variable("size(Reference)");
	                } else if (lookahead(TokenType.TEXT)) {
	                    Token<?> text = get();
	                    retval = new Variable("size(" + text.text.replace('/', '.') + ")");
	                } else {
	                    // primitive type
	                    get();
	                    if (!lookahead(TokenType.OPEN))
	                        error("expected '('");
	                    get();
	                    if (!lookahead(TokenType.TEXT))
	                        error("expected text");
	                    Token<?> text = get();
	                    if (!lookahead(TokenType.CLOSE))
	                        error("expected ')'");
	                    get();
	                    retval = new Variable("size(" + text.token + ")");
	                }
	                if (!lookahead(TokenType.COMMA))
	                    error("expected ','");
	                get();
	                if (!lookahead(TokenType.NUMBER))
	                    error("expected integer");
	                get();
	                if (!lookahead(TokenType.CLOSE))
	                    error("expected ')'");
	                get();
            	}
            } else {
                get();
            }
            if (!lookahead(TokenType.CLOSE))
                error("expected ')'");
            get();
            return retval; //new ReturnValue(retval);
        }
        if (lookahead(TokenType.VARIABLE)) {
            get();
            if (!lookahead(TokenType.OPEN))
                error("expected '('");
            get();
            //if (!lookahead(TokenType.TEXT))
            //error("expected text");
            //Token<String> text = (Token<String>) get();
            //Return retval = new ReturnExpression(text.token, 0, Return.PRIMARY_LEVEL);
            /*
            while (lookahead(TokenType.PLUS)) {
            	Token<?> t = get();
            	Return extra = expr();
            	retval = retval.plus(extra);
            }
            */
            Value retval = expr();
            if (!lookahead(TokenType.CLOSE))
                error("expected ')'");
            get();
            return retval;
        }
        if (lookahead(TokenType.MIN)) {
            get();
            if (lookaheadPrimary()) {
                Value pr = primary();
                return new Negation(pr);
            } else {
                error("expected primary after '-'");
            }
        }
        if (!lookahead(TokenType.NUMBER)) {
            error("did not recognise this text (expected variable or constant or something)");
        }
        Token<?> number = get();
        if(number.token instanceof Integer)
        	return new Number(((Integer) number.token).intValue());
        return null;
    }

    Value mulExpr() throws Exception {
        if (!lookaheadPrimary())
            error("primary expected");

        Value retval = primary();
        while (lookahead(TokenType.MUL)) {
            get();
            Value b = primary();
            retval = new MultiplicationOperation(retval, b);
            //System.out.println(aStr + " MUL " + b.toString(Return.TOP_LEVEL) + " = " + retval.toString(Return.TOP_LEVEL));
        }
        return retval;
    }

    Value expr() throws Exception {
        if (!lookaheadPrimary())
            error("primary expected");

        Value retval = mulExpr();
        while (lookahead(TokenType.PLUS) || lookahead(TokenType.MIN) || lookahead(TokenType.NUMBER)) {
            boolean min = lookahead(TokenType.MIN);
            if (lookahead(TokenType.NUMBER)) {
                Token<?> b = get();
                if(b.token instanceof Integer)
                retval = new AdditionOperation(retval, new Number((Integer) b.token));
                /*
                if (b.token < 0)
                	retval = retval.min(new ReturnValue(-b.token));
                else
                	retval = retval.plus(new ReturnValue(b.token));
                */
            } else {
                get();
                Value b = mulExpr();
                //retval = min ? retval.min(b) : retval.plus(b);
                retval = new AdditionOperation(retval, min ? new Negation(b) : b);
            }
            //System.out.println(aStr + " PLUS " + b.toString(Return.TOP_LEVEL) + " = " + retval.toString(Return.TOP_LEVEL));
        }
        return retval;
    }

    String start() {
        String retval = null;
        try {
            get();
            Value ret = expr();
            //System.out.println("result = " + ret.toString(Return.TOP_LEVEL) + "\n(current=" + current.toString() + ")");
            ret = process(ret);
            retval = ret.toString();
        } catch (Exception e) {
            return "fail!";
        }
        return retval;
    }
    
    public JCExpression convertToJML(Maker maker) {
    	JCExpression retval = null;
        try {
            get();
            Value ret = expr();
            //System.out.println("result = " + ret.toString(Return.TOP_LEVEL) + "\n(current=" + current.toString() + ")");
            ret = process(ret);
            retval = ret.translateToJML(maker);
        } catch (Exception e) {
            //System.out.println(e.toString());
            e.printStackTrace();
        }
        return retval;
    }

    Value powCompact(Value _ret) {
        if (!(_ret instanceof BinaryOperation))
            return _ret;
        //System.out.println("powCompact");
        BinaryOperation ret = (BinaryOperation)_ret;
        if (ret instanceof MultiplicationOperation) {
            // x^y * x -> x^(y+1)
            //System.out.println("powCompact: multiplication");
            Value _a = ret.firstComponent();
            Value b = ret.secondComponent();
            if (_a instanceof PowerOperation) {
                //System.out.println("powCompact: multiplication of pow");
                PowerOperation a = (PowerOperation)_a;
                //System.out.println("powCompact: a == b? " + a.firstComponent().toString() + "==" + b.toString());
                if (a.firstComponent().equals(b)) {
                    //System.out.println("powCompact: simplify");
                    AdditionOperation plus1 = new AdditionOperation(a.secondComponent(), new Number(1));
                    a.setSecondComponent(plus1);
                    ret = a;
                }
            }
        }
		ret.setFirstComponent(powCompact(ret.firstComponent()));
        ret.setSecondComponent(powCompact(ret.secondComponent()));
        return ret;
    }

    Value calculate(Value _ret) {
        if (!(_ret instanceof BinaryOperation))
            return _ret;
        BinaryOperation ret = (BinaryOperation)_ret;
        ret.setFirstComponent(calculate(ret.firstComponent()));
        ret.setSecondComponent(calculate(ret.secondComponent()));
        if (!(ret.firstComponent() instanceof Number) || !(ret.secondComponent() instanceof Number))
            return ret;
        Number a = (Number)ret.firstComponent();
        Number b = (Number)ret.secondComponent();
        if (ret instanceof AdditionOperation)
            return new Number(a.intValue()+b.intValue());
        if (ret instanceof MultiplicationOperation)
            return new Number(a.intValue()*b.intValue());
        if (ret instanceof PowerOperation)
            return new Number(new Double(Math.pow(a.intValue(),b.intValue())).intValue());
        if (ret instanceof MaxOperation)
        	return new Number(Math.max(a.intValue(), b.intValue()));
        return ret;
    }

    void collectBinary(Value prototype, Value current, boolean currentNegative, Vector<Number> concreteNumbers, Vector<Value> others, Vector<Value> othersNegative) {
        // this way, concrete numbers get calculated
        if (!prototype.getClass().equals(current.getClass())) {
            current = normalise(current);
        }
        if (current instanceof Number) { //.isConcrete()) {
            concreteNumbers.add((Number)current);
            return;
        }
        if (current instanceof Negation) {
            currentNegative = !currentNegative;
            current = ((Negation)current).value;
        }
        if (!prototype.getClass().equals(current.getClass())) {
            if (currentNegative)
                othersNegative.add(current);
            else
                others.add(current);
            return;
        }
        BinaryOperation binop = (BinaryOperation)current;
        collectBinary(prototype, binop.firstComponent(), currentNegative, concreteNumbers, others, othersNegative);
        collectBinary(prototype, binop.secondComponent(), currentNegative, concreteNumbers, others, othersNegative);
    }
    /*
    	Value _eliminateDuplicates(Vector<Value> all, int i, Value v) {
    		Value retval = null;
    		for (int j = i+1; j < all.size();) {
    			Value w = all.get(j);
    			boolean negative = w instanceof Negation;
    			if (negative)
    				w = ((Negation)w).value;
    			if (v.equals(w)) {
    				if (retval != null)
    					retval = new AdditionOperation(retval, new Number(negative ? -1 : 1));
    				else
    					retval = new Number(negative ? -1 : 1);
    				all.remove(j);
    			} else if (w instanceof MultiplicationOperation) {
    				// misschien een van de componenten?
    				MultiplicationOperation m = (MultiplicationOperation)w;
    				if (v.equals(m.firstComponent())) {
    					Value snd = m.secondComponent();
    					if (negative)
    						snd = new Negation(snd);
    					if (retval != null) {
    						retval = new AdditionOperation(retval, snd);
    					} else
    						retval = snd;
    					all.remove(j);
    				} else if (v.equals(m.secondComponent())) {
    					Value fst = m.firstComponent();
    					if (negative)
    						fst = new Negation(fst);
    					if (retval != null) {
    						retval = new AdditionOperation(retval, fst);
    					} else
    						retval = fst;
    					all.remove(j);
    				} else {
    					j++;
    				}
    			} else {
    				j++;
    			}
    		}
    		return retval;
    	}
    	boolean eliminateDuplicates(Vector<Value> all) {
    		boolean retval = false;
    		for (int i = 0; i < all.size(); i++) {
    			Value v = all.get(i);
    			Value count = null;
    			if (v instanceof Negation) {
    				v = ((Negation)v).value;
    				count = new Number(-1);
    			} else
    				count = new Number(1);
    			boolean replace = false;
    			if (v instanceof MultiplicationOperation) {
    				MultiplicationOperation mul = (MultiplicationOperation)v;
    				if (mul.firstComponent() instanceof Number) {
    					count = new MultiplicationOperation(mul.firstComponent(), count);
    					Value result = _eliminateDuplicates(all, i, mul.secondComponent());
    					if (result != null) {
    						replace = true;
    						count = new AdditionOperation(count, result);
    						v = mul.secondComponent();
    					}
    				} else {
    					count = new MultiplicationOperation(mul.secondComponent(), count);
    					Value result = _eliminateDuplicates(all, i, mul.firstComponent());
    					if (result != null) {
    						replace = true;
    						count = new AdditionOperation(count, result);
    						v = mul.firstComponent();
    					}
    				}
    			} else {
    				Value result = _eliminateDuplicates(all, i, v);
    				if (result != null) {
    					replace = true;
    					count = new AdditionOperation(count, result);
    				}
    			}
    			if (replace) {
    				count = normalise(count);
    				if (v instanceof Negation) {
    					v = new MultiplicationOperation(new Negation(count), ((Negation)v).value);
    				} else
    					v = new MultiplicationOperation(count, v);
    				v = normalise(v);
    				all.setElementAt(v, i);
    				retval = true;
    			}
    		}
    		return retval;
    	}
    */
    Value _eliminateDuplicatesInAddition(Vector<Value> all, int i, Value v) {
        Value retval = null;
        for (int j = i+1; j < all.size(); j++) {
            Value w = all.get(j);
            //System.out.println("check if term " + j + "='"+w.toString()+"' has a term of '" + v.toString() + "'");
            /*
            Value count = w.countTerm(v, true);
            boolean replacable = additionLike;
            if (count == null) {
            	replacable = false;
            	count = w.countTerm(v, false);
            }
            //if (w.hasTerm(v)) {
            if (count != null) {
            	System.out.println("positive, namely " + count.toString());
            	if (retval != null)
            		retval = new AdditionOperation(retval, count);
            	else
            		retval = count;
            	if (replacable) {
            		all.remove(j);
            		j--;
            	} else {
            		Value replacement = new AdditionOperation(w, new Negation(new MultiplicationOperation(count, v)));
            		//System.err.println("Normalising replacement: " + replacement.toString());
            		//replacement = simplify(replacement);
            		if (replacement instanceof Number && ((Number)replacement).intValue() == 0) {
            			all.remove(j);
            			j--;
            		} else {
            			all.setElementAt(replacement, j);
            		}
            	}
            }
            */
            Value count = w.countTerm(v, false);
            if (count != null) {
                if (retval != null)
                    retval = new AdditionOperation(retval, count);
                else
                    retval = count;
                Value wo = w.without(v);
                if (wo != null) {
                    //System.err.println("-> replacing '" + w + "' without '" + v + "' with: '" + wo + "'");
                    all.setElementAt(wo, j);
                } else {
                    //System.err.println("-> removing '" + w + "' without '" + v + "'");
                    all.remove(j);
                    j--;
                }
            }
        }
        //System.err.println("-> " + retval);
        return retval;
    }
    boolean eliminateDuplicatesInAddition(Vector<Value> all) {
        //System.err.println();
        //System.err.println("Eliminate duplicates");

        boolean retval = false;
        for (int i = 0; i < all.size(); i++) {
            Value v = all.get(i);
            //System.err.println("Eliminate step of term " + i + ": " + v.toString());
            Vector<Value> terms = new Vector<Value>();
            v.allTerms(terms);

            for (Value term : terms) {
                // only for variables
                //if (!(term instanceof Variable) && !(term instanceof Number))
                //	continue;
                //System.err.println("\tstep; finding duplicates of term: " + term.toString());
                Value count = v.countTerm(term, true);
                if (count == null)
                    continue;
                //System.err.println("\tresults in " + count.toString());
                boolean replace = false;
                Value result = _eliminateDuplicatesInAddition(all, i, term);
                if (result != null) {
                    replace = true;
                    count = new AdditionOperation(count, result);
                }
                if (replace) {
                    //count = normalise(count);
                    v = new MultiplicationOperation(count, term);
                    v = normalise(v);
                    //System.err.println("Replacing '" + all.get(i) + "' with: " + v.toString());
                    all.setElementAt(v, i);
                    retval = true;
                    break;
                }
            }
        }
        //System.err.println("Finished eliminating duplicates");
        return retval;
    }
    boolean eliminateDuplicatesInMultiplication(Vector<Value> all) {
        //System.err.println();
        //System.err.println("Eliminate duplicates in multiplication");

        boolean retval = false;
        for (int i = 0; i < all.size(); i++) {
            Value v = all.get(i);
            //System.err.println("Eliminate step of term " + i + ": " + v.toString());
            Vector<Value> terms = new Vector<Value>();
            v.allTerms(terms);

            for (Value term : terms) {
                // only for variables
                //if (!(term instanceof Variable) && !(term instanceof Number))
                //	continue;
                //System.err.println("\tstep; finding duplicates of term: " + term.toString());
                if (!v.equals(term))
                    continue;
                Value count = new Number(1); //v.countTerm(term, true);
                //System.err.println("\tresults in " + count.toString());
                boolean replace = false;
                for (int j = i+1; j < all.size(); j++) {
                    Value c2 = all.get(j).timesTerm(term);
                    if (c2 != null) {
                        count = new AdditionOperation(count, c2);
                        Value wo = all.get(j).without(term);
                        if (wo == null) {
                            //System.err.println("Removing " + all.get(i) + " count=" + c2);
                            all.remove(j);
                            j--;
                        } else {
                            //System.err.println("Replacing (loop) " + all.get(j) + " with: " + wo.toString() + " count=" + c2);
                            all.setElementAt(wo, j);
                        }
                        replace = true;

                    }
                    /*
                    if (term.equals(all.get(j))) {
                    	count = new AdditionOperation(count, new Number(1));
                    	all.remove(j);
                    	j--;
                    	replace = true;
                    }
                    */
                    
                }
                if (replace) {
                    v = new PowerOperation(term, count);
                    v = normalise(v);
                    //System.err.println("Replacing (mul) " + all.get(i) + " with: " + v.toString());
                    all.setElementAt(v, i);
                    retval = true;
                    break;
                }
            }
        }
        for (int i = 0; i < all.size(); i++) {
            Value _a = all.get(i);
            if (_a instanceof PowerOperation) {
                PowerOperation a = (PowerOperation)_a;
                for (int j = i+1; j < all.size(); j++) {
                    Value b = all.get(j);
                    if (a.firstComponent().equals(b)) {
                    	AdditionOperation plus1 = new AdditionOperation(a.secondComponent(), new Number(1));
                        a.setSecondComponent(plus1);
                        all.remove(j);
                        j--;
                        retval = true;
                    } else if (b instanceof PowerOperation) {
                    	PowerOperation powb = (PowerOperation)b;
                    	if (a.firstComponent().equals(powb.firstComponent())) {
                    		AdditionOperation plus1 = new AdditionOperation(a.secondComponent(), powb.secondComponent());
                            a.setSecondComponent(plus1);
                            all.remove(j);
                            j--;
                            retval = true;
                    	}
                    }
                }
            }
        }
        //System.err.println("Finished eliminating duplicates");
        return retval;
    }
    Value normalise(Value _ret) {
        boolean returnNegative = false;
        if (!(_ret instanceof BinaryOperation)) {
            if (_ret instanceof Negation) {
                Negation ret = (Negation)_ret;
                ret.value = normalise(ret.value);
                if (ret.value instanceof Negation) {
                    ret = (Negation)(ret.value);
                    _ret = ret.value;
                } else if (ret.value instanceof Number) {
                    Number n = (Number)ret.value;
                    _ret = new Number(-n.intValue());
                }
            }
            return _ret;
        }
        //System.err.println("Normalise: " + _ret.toString());
        BinaryOperation ret = (BinaryOperation)_ret;
        if (!ret.isSymmetrical()) {
            ret.a = normalise(ret.a);
            ret.b = normalise(ret.b);
            if (ret instanceof PowerOperation) {
                PowerOperation pow = (PowerOperation)ret;
                if (pow.isConcrete()) {
                    ret.setFirstComponent(normalise(pow.firstComponent()));
                    ret.setSecondComponent(normalise(pow.secondComponent()));
                    Number a = (Number)pow.firstComponent();
                    Number b = (Number)pow.secondComponent();
                    return new Number(new Double(Math.pow(a.intValue(),b.intValue())).intValue());
                } else if (pow.secondComponent() instanceof Number && ((Number)pow.secondComponent()).intValue() == 0) {
                    return new Number(1);
                } else if (pow.secondComponent() instanceof Number && ((Number)pow.secondComponent()).intValue() == 1) {
                    return pow.firstComponent();
                } else if (pow.firstComponent() instanceof PowerOperation) {
                    // in pow; n^2^4 => n^8
                	PowerOperation pow2 = (PowerOperation)pow.firstComponent();
                	ret.setFirstComponent(pow2.firstComponent());
                	ret.setSecondComponent(new MultiplicationOperation(pow2.secondComponent(), pow.secondComponent()));
                }
            }
            return ret;
        }
        if(ret instanceof MaxOperation){
        	MaxOperation max = (MaxOperation) ret;
        	max.a = normalise(max.a);
        	max.b = normalise(max.b);
        	if(max.isConcrete())
        		return new Number(Math.max(((Number) max.a).intValue(),((Number) max.b).intValue()));
        	else
        		return max;
        }
        Vector<Number> concreteNumbers = new Vector<Number>();
        Vector<Value> allPos = new Vector<Value>();
        Vector<Value> allNeg = new Vector<Value>();

        // collect all the arguments of this type
        collectBinary(ret, ret, false, concreteNumbers, allPos, allNeg);

        // combine concrete numbers
        if (concreteNumbers.size() > 0) {
            int current = concreteNumbers.remove(0).intValue();
            for (Number v : concreteNumbers) {
                if (ret instanceof AdditionOperation)
                    current += v.intValue();
                else if (ret instanceof MultiplicationOperation)
                    current *= v.intValue();
            }
            if (ret instanceof MultiplicationOperation && current == 0)
                return new Number(0);
            if (ret instanceof MultiplicationOperation && current == -1) {
                returnNegative = true;
            } else if (!(ret instanceof AdditionOperation && current == 0) && !(ret instanceof MultiplicationOperation && current == 1))
                allPos.add(new Number(current));
        }
        /*
        if (ret instanceof AdditionOperation)
        	return new Number(a.intValue()+b.intValue());
        if (ret instanceof MultiplicationOperation)
        	return new Number(a.intValue()*b.intValue());
        if (ret instanceof PowerOperation)
        	return new Number(new Double(Math.pow(a.intValue(),b.intValue())).intValue());
        */
        // in mul; -n*-m => n*m
        if (ret instanceof MultiplicationOperation) {
            while (allNeg.size() >= 2) {
                allPos.add(allNeg.remove(0));
                allPos.add(allNeg.remove(0));
            }
        }
        // in add; eliminate n-n
        if (ret instanceof AdditionOperation) {
            for (int i = 0; i < allPos.size(); ) {
                Value v = allPos.get(i);
                if (allNeg.contains(v)) {
                    allPos.remove(v);
                    allNeg.remove(v);
                } else {
                    i++;
                }
            }
        }

        // TODO: in mul; simplify duplicate variables (n*n => n^2)
        // TODO: in add; get common variables outside the term, like 5*n + 5*m + 5*o => 5*(n+m+o)
        // TODO: integrate the powCompact operation here, as it should work on all pairs.
        // TODO: in pow: x^y*x => x^(y+1)
        // TODO: add y*x+x => (y+1)*x

        // rebuild the normalised tree
        Value retval = null;
        Vector<Value> all = allPos;
        for (Value v : allNeg) {
            all.add(new Negation(v));
        }
        // TODO: in add; simplify duplicate variables (n+n => n*2)
        if (ret instanceof AdditionOperation) {
            eliminateDuplicatesInAddition(all);
        }
        if (ret instanceof MultiplicationOperation) {
            eliminateDuplicatesInMultiplication(all);
        }
        if (all.size() > 0) {
            retval = all.remove(0);
            while (all.size() > 1) {
                Value next = all.remove(0);
                retval = ret.create(retval, next);
            }
            if (all.size() == 1) {
                Value last = all.remove(0);
                // for multiplication put the concrete number in front, so the output will be 3*n, otherwise retain the original order
                if (!(last instanceof Number) || !(ret instanceof MultiplicationOperation)) {
                    retval = ret.create(retval, last);
                } else {
                    retval = ret.create(last, retval);
                }
            }
        } else {
            // nothing, so 0
            retval = new Number(0);
        }

        //System.err.println("normalise: " + _ret.toString() + " -> " + retval.toString());
        //if (extra)
        //	retval = normalise(retval);
        // done
        if (returnNegative)
            return new Negation(retval);
        return retval;
    }

    Value process(Value ret) {
        //System.out.println("start: " + ret.toString());
        ret = normalise(ret);
        //System.out.println("after normalise: " + ret.toString());
        //ret = powCompact(ret);
        ret = normalise(ret);
        //System.out.println("after powCompact: " + ret.toString());
        //ret = calculate(ret);
        return ret;
    }

    public static void main(String[] args) throws Exception {
        String[] testCases = {
            // basic stuff
            "2+2", "4",
            "2-1", "1",
            "2*2", "4",
            "pow(2,3)", "8",
            "2*(2+2)", "8",
            "1+2*3", "7",
            "3*3-2*2", "5",

            // variable stuff
            "nat(n)*2", "2*n",
            "nat(n)+3", "n+3",
            "nat(x)*nat(y)", "x*y",
            "pow(2,nat(n))", "2^n",
            "nat(n)*2*3", "6*n",
            "3*2*nat(n)", "6*n",
            "3*nat(n)*2", "6*n",
            "-nat(n)", "-n",
            "-(nat(n))", "-n",
            "-(-(nat(n)))", "n",
            "-(-(nat(n))+0)", "n",
            "-(-(nat(n))*1)", "n",
            "-nat(n)*-nat(x)", "n*x",
            "(nat(a)+nat(b))-(nat(x)+nat(y))", "a+b-x-y",
            "(3*nat(n))-2", "3*n-2",
            "2-(3*nat(n))", "2-3*n",
            "-(3*nat(n))+2", "2-3*n",
            "y*x+x", "(y+1)*x",
            "3*nat(n)-(2*nat(n))-nat(n)+nat(m)", "m",
            "-1*n", "-n",
            "x*y+(x+1)*z", "(y+z)*x+z",
            "x+x*x", "(x+1)*x",
            "3*n+2*n+n+m", "6*n+m",
            "(x+2)*a+(x+2)*b+(x+2)*d", "(a+b+d)*(x+2)",

            // [..] * 1 or [..] + 0 or [..]^0
            "nat(n)+0", "n",
            "nat(n)*1", "n",
            "nat(x)*(nat(n)+0)", "x*n",
            "nat(x)*(nat(n)*1)", "x*n",
            "pow(nat(n),0)", "1",

            // [..] * 0 or n - n
            "nat(n)+nat(m)-nat(m)", "n",
            "nat(n)-nat(n)", "0",
            "nat(n)*0", "0",

            // other
            "nat(n)+nat(n)", "2*n",
            "-nat(n)-nat(n)", "-2*n",
            "2*nat(n)-nat(n)", "n",
            "2*nat(n)+nat(n)", "3*n",
            "2*nat(n)*nat(k)+nat(n)", "(2*k+1)*n",
            "nat(n)*nat(n)", "n^2",
            "pow(nat(n),2)*pow(nat(n),2)", "n^4",
            "pow(nat(n),3)*pow(nat(n),2)", "n^5",
            "pow(pow(nat(n),2),3)*pow(nat(n),2)", "n^8",
            "pow(n,2)*(pow(pow(n,2),3)+n)", "n^3*(n^5+1)",
            "2*nat(n)+2*nat(m)+2*nat(o)", "2*(n+m+o)",
            //"(nat(a)+nat(b))*((nat(c)+nat(d))", "a*c+a*d+b*c+b*d"
            "pow(n,2)+4*n+3", "(n+4)*n+3", // (n+3)*(n+1)
            "pow(n,2)+3*n+3*k+n*k", "(n+k+3)*n+3*k", // (n+3)(n+k) 
            "pow(n,nat(a-1))*n", "n^a",
            "pow(n,nat(a-1))*k*n", "n^a*k",
            
            "max([0,0])", "0",
            "max([1,0])", "1",
            "max([0,1])", "1",
            "max([6,3])", "6",
            "max([3,6])", "6",
            "max([2*3+1,pow(2,3)])", "8",
            "max([max([2+3,pow(2,3)]),2*3])", "8",
        };
        //System.out.println(new JaSB("3*3-2 *2", ".", true).start());
        
        if (args.length > 0) {
            System.out.println(new JaSB(args[0], ".", true, "jamaica",false).start());
        } else {
			API api = new API();
			// Make sure we have a context
			api.parseString("temp", "class A{}");
            int success = 0;
            int failed = 0;
            for (int i = 0; i < testCases.length; i = i+2) {
                String expr = testCases[i];
                String result = new JaSB(expr, ".", true, "jamaica",false).start();
                String expected = testCases[i+1];
                if (!expected.equals(result)) {
                    System.out.println("[FAILED] " + expr + "\n      -> " + result + "\n      != " + expected);
                    failed++;
                } else {
                    System.out.println("[OK] " + expr + "\n  -> " + result);
                    success++;
                }
                //JCExpression jml = new JaSB(expr, ".", true).convertToJML(maker);
                //System.out.println("JML: " + jml.toString());
            }
            System.out.println(failed + " failed tests of the " + (failed + success) + " tests.");
        }
    }

}
