package resanaplugin.loopbounds;

import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.widgets.*;

/**
 * 
 * @author Liviu Varga
 * @author Rody Kersten
 * @author Nick Gubbels
 *
 */

public class InputOptionsDialog extends Dialog {
	private Spinner degreeSpinner;
	Button ifsplit, standardJMLButton, overwrite;
	Button stepAuto, stepEqual;
	Button refactoringButton;
	Spinner stepSpinner;
	Spinner nodestartpointSpinner, nodeminvalSpinner, nodemaxvalSpinner;

	/**
	 * @param parent
	 */
	public InputOptionsDialog(Shell parent) {
		super(parent);
	}

	/**
	 * @param parent
	 * @param style
	 */
	public InputOptionsDialog(Shell parent, int style) {
		super(parent, style);
	}

	/**
	 * Makes the dialog with the necessary options
	 * 
	 * @return
	 */
	private void CreateOptions(Shell sh) {	
		GridData griddata;

		// DEGREE
		Group degreeGroup = new Group(sh, SWT.NONE);
		degreeGroup.setLayout(new GridLayout (2, false));
		degreeGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		degreeGroup.setText("Degree");	
		
		Label degreeLabel  = new Label(degreeGroup, SWT.RIGHT);
		degreeLabel.setText("Expected degree: ");
		griddata = new GridData(115, 25);
		griddata.horizontalAlignment = SWT.RIGHT;
		griddata.verticalAlignment = SWT.CENTER;
		griddata.grabExcessHorizontalSpace = false;
		griddata.grabExcessVerticalSpace = false;
		degreeLabel.setLayoutData(griddata);
		
		degreeSpinner = new Spinner(degreeGroup, SWT.NONE);
		griddata = new GridData(40, 20);
		griddata.horizontalAlignment = SWT.LEFT;
		griddata.verticalAlignment = SWT.CENTER;
		griddata.grabExcessHorizontalSpace = false;
		griddata.grabExcessVerticalSpace = false;
		degreeSpinner.setLayoutData(griddata);
		degreeSpinner.setMinimum(0);
		degreeSpinner.setMaximum(99);
		degreeSpinner.setSelection(1);
		
		
		
		// BRANCHING
		Group branchingGroup = new Group(sh, SWT.NONE);
		branchingGroup.setLayout(new GridLayout (1, false));
		branchingGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		branchingGroup.setText("Branch splitting");
		
		ifsplit = new Button(branchingGroup, SWT.CHECK);
		ifsplit.setText("Apply branch-splitting");
		griddata = new GridData();
		griddata.horizontalSpan = 2;
		ifsplit.setLayoutData(griddata);
		
		
		// STEPS
		Group stepsGroup = new Group(sh, SWT.NONE);
		stepsGroup.setLayout(new GridLayout (2, false));
		stepsGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		stepsGroup.setText("Steps");

		stepAuto = new Button(stepsGroup, SWT.RADIO);
		stepAuto.setText("Infer steps");
		stepAuto.addListener (SWT.Selection, new Listener () {
			public void handleEvent (Event e) {
				if (stepAuto.getSelection()) {
					stepSpinner.setEnabled(false);
				}
			}
		});
		new Label(stepsGroup, SWT.RIGHT); //empty label
		stepEqual = new Button(stepsGroup, SWT.RADIO);
		stepEqual.setText("All steps are: ");
		stepEqual.addListener (SWT.Selection, new Listener () {
			public void handleEvent (Event e) {
				if (stepEqual.getSelection()) {
					stepSpinner.setEnabled(true);
				}
			}
		});
		stepSpinner = new Spinner(stepsGroup, SWT.NONE);
		griddata = new GridData(40, 20);
		griddata.horizontalAlignment = SWT.LEFT;
		griddata.verticalAlignment = SWT.CENTER;
		griddata.grabExcessHorizontalSpace = false;
		griddata.grabExcessVerticalSpace = false;
		stepSpinner.setLayoutData(griddata);
		stepSpinner.setMinimum(1);
		stepSpinner.setMaximum(10000);
		stepSpinner.setSelection(1);
		stepAuto.setSelection(true);
		stepSpinner.setEnabled(false);
		
		
		// NODE SEARCH
		Group nodesearchGroup = new Group(sh, SWT.NONE);
		nodesearchGroup.setLayout(new GridLayout (1, false));
		griddata = new GridData(SWT.FILL, SWT.FILL, false, false);
		nodesearchGroup.setLayoutData(griddata);
		nodesearchGroup.setText("Node search");
		
		Label nodestartpointLabel  = new Label(nodesearchGroup, SWT.RIGHT);
		nodestartpointLabel.setText("Node-search start point: ");
		nodestartpointSpinner = new Spinner(nodesearchGroup, SWT.NONE);
		griddata = new GridData(40, 20);
		griddata.horizontalAlignment = SWT.LEFT;
		griddata.verticalAlignment = SWT.CENTER;
		griddata.grabExcessHorizontalSpace = true;
		griddata.grabExcessVerticalSpace = false;
		nodestartpointSpinner.setLayoutData(griddata);
		nodestartpointSpinner.setMinimum(Integer.MIN_VALUE);
		nodestartpointSpinner.setMaximum(Integer.MAX_VALUE);
		nodestartpointSpinner.setSelection(0);
		
		Label nodeminvalLabel  = new Label(nodesearchGroup, SWT.RIGHT);
		nodeminvalLabel.setText("Node-search minimum: ");
		nodeminvalSpinner = new Spinner(nodesearchGroup, SWT.NONE);
		griddata = new GridData(40, 20);
		griddata.horizontalAlignment = SWT.LEFT;
		griddata.verticalAlignment = SWT.CENTER;
		griddata.grabExcessHorizontalSpace = false;
		griddata.grabExcessVerticalSpace = false;
		nodeminvalSpinner.setLayoutData(griddata);
		nodeminvalSpinner.setMinimum(Integer.MIN_VALUE);
		nodeminvalSpinner.setMaximum(Integer.MAX_VALUE);
		nodeminvalSpinner.setSelection(-1000);
		nodeminvalSpinner.addListener (SWT.Modify, new Listener () {
			public void handleEvent (Event e) {
				nodestartpointSpinner.setMinimum(nodeminvalSpinner.getSelection());
				nodemaxvalSpinner.setMinimum(nodeminvalSpinner.getSelection());
			}
		});
		
		Label nodemaxvalLabel  = new Label(nodesearchGroup, SWT.RIGHT);
		nodemaxvalLabel.setText("Node-search maximum: ");
		nodemaxvalSpinner = new Spinner(nodesearchGroup, SWT.NONE);
		griddata = new GridData(40, 20);
		griddata.horizontalAlignment = SWT.LEFT;
		griddata.verticalAlignment = SWT.CENTER;
		griddata.grabExcessHorizontalSpace = true;
		griddata.grabExcessVerticalSpace = false;
		nodemaxvalSpinner.setLayoutData(griddata);
		nodemaxvalSpinner.setMinimum(Integer.MIN_VALUE);
		nodemaxvalSpinner.setMaximum(Integer.MAX_VALUE);
		nodemaxvalSpinner.setSelection(1000);
		nodemaxvalSpinner.addListener (SWT.Modify, new Listener () {
			public void handleEvent (Event e) {
				nodestartpointSpinner.setMaximum(nodemaxvalSpinner.getSelection());
				nodeminvalSpinner.setMaximum(nodemaxvalSpinner.getSelection());
			}
		});

		// ANNOTATIONS
		Group annotationsGroup = new Group(sh, SWT.NONE);
		annotationsGroup.setLayout(new GridLayout (1, false));
		annotationsGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		annotationsGroup.setText("Annotations");
		standardJMLButton = new Button(annotationsGroup, SWT.CHECK);
		standardJMLButton.setText("Output JML annotations as given in JML specification (not verifiable by KeY)");
		overwrite = new Button(annotationsGroup, SWT.CHECK);
		overwrite.setText("Overwrite existing decreases-clause");
		
		// AUTOMATIC REFACTORING
		Group refactoringGroup = new Group(sh, SWT.NONE);
		refactoringGroup.setLayout(new GridLayout(1,false));
		refactoringGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		refactoringGroup.setText("Automatic refactoring");
		
		refactoringButton = new Button(refactoringGroup, SWT.CHECK);
		refactoringButton.setText("Use automatic refactoring");
		refactoringButton.setSelection(true);
	}
	
	/**
	 * Activates the input dialog window for options like degree, step size, if splitting, find step
	 * 
	 * @return The selected options (as an Options object)
	 */
	public Options open() {
		final Options result = new Options();
		Shell parent = getParent();
		final Shell shell = new Shell(parent, SWT.TITLE | SWT.BORDER
				| SWT.APPLICATION_MODAL);
		shell.setText("ResAna options");
		shell.setLayout(new GridLayout(2, true));
		shell.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

		CreateOptions(shell);

		new Label(shell, SWT.RIGHT); //empty label to move buttons to the right
		
		Composite buttons = new Composite(shell, SWT.FILL);
		buttons.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		FillLayout layout = new FillLayout();
		buttons.setLayout(layout);
		
		final Button buttonCancel = new Button(buttons, SWT.PUSH);
		buttonCancel.setText("Cancel");
		buttonCancel.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				shell.dispose();
			}
		});
		
		final Button buttonOK = new Button(buttons, SWT.PUSH);
		buttonOK.setText("Ok");
		buttonOK.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				result.ok = true;
				result.degree = degreeSpinner.getSelection();
				result.step = stepAuto.getSelection() ? -1 : stepSpinner.getSelection();
				result.branchSplitting = ifsplit.getSelection();
				result.correctJML = standardJMLButton.getSelection();
				result.overwrite = overwrite.getSelection();
				result.nodeStartPoint = nodestartpointSpinner.getSelection();
				result.nodeMinVal = nodeminvalSpinner.getSelection();
				result.nodeMaxVal = nodemaxvalSpinner.getSelection();
				result.refactoring = refactoringButton.getSelection();
				shell.dispose();
			}
		});
		
		shell.addListener(SWT.Traverse, new Listener() {
			public void handleEvent(Event event) {
				if (event.detail == SWT.TRAVERSE_ESCAPE)
					event.doit = false;
			}
		});

		shell.pack();
		shell.open();

		Display display = parent.getDisplay();
		while (!shell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}
		
		return result;
	}
}